import type { DocumentReference, Firestore } from '@google-cloud/firestore';
import type { Collection } from './collection';
/**
 * Deep equality algorithm based on `_.isEqual()` with special handling
 * of objects that have their own `isEqual()` method, such as `Reference`.
 */
export declare function isEqualHandlingRef(a: any, b: any): boolean;
/**
 * The shape of references as stored in Firestore.
 */
export declare type ReferenceShape<T extends object> = DocumentReference<T> | FlatReferenceShape<T> | MorphReferenceShape<T>;
export interface FlatReferenceShape<T extends object> {
    ref: DocumentReference<{
        [id: string]: T;
    }>;
    id: string;
}
export declare type MorphReferenceShape<T extends object> = NormalMorphReferenceShape<T> | FlatMorphReferenceShape<T>;
export interface NormalMorphReferenceShape<T extends object> {
    ref: DocumentReference<T>;
    filter: string | null;
}
export interface FlatMorphReferenceShape<T extends object> extends FlatReferenceShape<T> {
    filter: string | null;
}
export interface Snapshot<T extends object> {
    data(): T | undefined;
    ref: Reference<T>;
    id: string;
    exists: boolean;
}
export interface SetOpts {
    /**
     * Indicates whether relation links should be updated on
     * other related documents. This can extra reads, queries
     * and writes to multiple documents.
     *
     * If not updated, reference links will get into invalid states,
     * so it should be done unless you know what you're doing.
     *
     * Defaults to `true`.
     */
    updateRelations?: boolean;
    /**
     * Whether or not the `onChange(...)` hook will be run. If not provided,
     * it defaults to the value of `updateRelations` (which defaults to `true`).
     *
     * This default behaviour means that it will run for all explicit changes, but not
     * for implicit changes internally caused to related documents when relations are updated.
     */
    runOnChangeHook?: boolean;
}
/**
 * Common interface for normal, flattened, and polymorphic references.
 * References perform coercion on input data according to the model
 * schema that they belong to.
 */
export declare abstract class Reference<T extends object> {
    abstract readonly parent: Collection<T>;
    abstract readonly id: string;
    abstract readonly path: string;
    abstract readonly firestore: Firestore;
    abstract delete(opts?: SetOpts): Promise<void>;
    /**
     * @returns The coerced data
     */
    abstract create(data: T, opts?: SetOpts): Promise<T>;
    abstract create(data: Partial<T>, opts?: SetOpts): Promise<Partial<T>>;
    /**
     * @returns The coerced data
     */
    abstract update(data: T, opts?: SetOpts): Promise<T>;
    abstract update(data: Partial<T>, opts?: SetOpts): Promise<Partial<T>>;
    abstract get(): Promise<Snapshot<T>>;
    abstract isEqual(other: any): boolean;
    /**
     * Allow serialising to JSON.
     */
    abstract toJSON(): any;
    /**
     * Returns a value that can be serialised to Firestore.
     */
    abstract toFirestoreValue(): ReferenceShape<T>;
    toString(): string;
}
