"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.buildIndexers = exports.generateMetadataForComponent = exports.updateComponentsMetadata = exports.doesComponentRequireMetadata = void 0;
const _ = require("lodash");
const reference_1 = require("../db/reference");
const components_1 = require("./components");
function doesComponentRequireMetadata(attr) {
    return (attr.type === 'dynamiczone')
        || ((attr.type === 'component') && (attr.repeatable === true));
}
exports.doesComponentRequireMetadata = doesComponentRequireMetadata;
function updateComponentsMetadata(model, data, output = data) {
    if (model.isComponent) {
        return;
    }
    for (const parentAlias of model.componentKeys) {
        // Don't overwrite metadata with empty map if the value 
        // doesn't exist because this could be a partial update
        if (!_.has(data, parentAlias)) {
            continue;
        }
        Object.assign(output, generateMetadataForComponent(model, parentAlias, data));
    }
}
exports.updateComponentsMetadata = updateComponentsMetadata;
function generateMetadataForComponent(model, parentAlias, data) {
    const parentAttr = model.attributes[parentAlias];
    if (doesComponentRequireMetadata(parentAttr)) {
        const metaField = model.getMetadataMapKey(parentAlias);
        // Initialise the map will null for all known keys
        const meta = {};
        const componentModels = parentAttr.component ? [parentAttr.component] : (parentAttr.components || []);
        for (const modelName of componentModels) {
            const { indexers = [] } = (0, components_1.getComponentModel)(modelName);
            for (const info of indexers) {
                for (const key of Object.keys(info.indexers)) {
                    meta[key] = null;
                }
            }
        }
        // Make an array containing the value of this attribute from all the components in the array
        // If the value itself is an array then is is concatenated/flattened
        const components = _.castArray(_.get(data, parentAlias) || []);
        for (const component of components) {
            const componentModel = (0, components_1.getComponentModel)(model, parentAlias, component);
            if (!componentModel.indexers) {
                continue;
            }
            for (const { alias, indexers } of componentModel.indexers) {
                const values = _.castArray(_.get(component, alias, []));
                for (const key of Object.keys(indexers)) {
                    const arr = meta[key] = meta[key] || [];
                    const indexer = indexers[key];
                    for (let value of values) {
                        const result = indexer(value, component);
                        // Only add if the element doesn't already exist
                        // and is not undefined
                        if ((result !== undefined)
                            && (!arr.some(v => (0, reference_1.isEqualHandlingRef)(v, result)))) {
                            arr.push(result);
                        }
                    }
                }
            }
        }
        // Ensure all empty indexes are null
        for (const key of Object.keys(meta)) {
            const arr = meta[key];
            if (!arr || !arr.length) {
                meta[key] = null;
            }
            // Add the empty indicator
            meta[`${key}$empty`] = ((arr === null || arr === void 0 ? void 0 : arr.length) === 0);
        }
        // Only assign meta if there are any keys
        if (Object.keys(meta).length) {
            return { [metaField]: meta };
        }
        else {
            return {};
        }
    }
    return undefined;
}
exports.generateMetadataForComponent = generateMetadataForComponent;
/**
 * Build indexers for all the indexed attributes
 * in a component model.
 */
function buildIndexers(model) {
    if (model.modelType !== 'component') {
        return undefined;
    }
    const infos = [];
    for (const alias of Object.keys(model.attributes)) {
        const attr = model.attributes[alias];
        const isRelation = attr.model || attr.collection;
        if (isRelation || attr.index) {
            let defaultIndexer;
            let indexers;
            if (typeof attr.index === 'object') {
                indexers = {};
                for (const key of Object.keys(attr.index)) {
                    const indexer = attr.index[key];
                    if (indexer) {
                        if (typeof indexer === 'function') {
                            indexers[key] = indexer;
                        }
                        else {
                            indexers[key] = defaultIndexerFn;
                            if (!defaultIndexer) {
                                defaultIndexer = key;
                            }
                        }
                    }
                }
                // Ensure there is a default indexer for relation types
                if (isRelation && !defaultIndexer) {
                    defaultIndexer = alias;
                    indexers[alias] = defaultIndexerFn;
                }
            }
            else {
                const key = (typeof attr.index === 'string') ? attr.index : alias;
                defaultIndexer = key;
                indexers = {
                    [key]: defaultIndexerFn,
                };
            }
            // Delete index info from the original attribute definition
            // no longer needed and it clutters API metadata etc
            delete attr.index;
            // If this indexer is for the primary key and only the indexer is defined
            // then this is a special case and we remove the attribute once
            // the indexers have been absorbed
            if ((alias === model.primaryKey) && !Object.keys(attr).length) {
                delete model.attributes[alias];
            }
            infos.push({
                alias,
                attr,
                defaultIndexer,
                indexers,
            });
        }
    }
    return infos;
}
exports.buildIndexers = buildIndexers;
function defaultIndexerFn(value) {
    return value;
}
