"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.RelationHandler = void 0;
const _ = require("lodash");
const status_error_1 = require("./status-error");
const morph_reference_1 = require("../db/morph-reference");
const field_operation_1 = require("../db/field-operation");
const reference_1 = require("../db/reference");
const normal_reference_1 = require("../db/normal-reference");
const deep_reference_1 = require("../db/deep-reference");
const map_not_null_1 = require("./map-not-null");
class RelationHandler {
    constructor(thisEnd, otherEnds) {
        this.thisEnd = thisEnd;
        this.otherEnds = otherEnds;
        if (!thisEnd.attr && !otherEnds.some(e => e.attr)) {
            throw new Error('Relation does not have any dominant ends defined');
        }
        if (otherEnds.some(e => e.model.isComponent && (!e.attr || thisEnd.attr))) {
            throw new Error('Relation cannot have a dominant reference to a component');
        }
        // Virtual collections are probably transient/unstable unless `hasStableIds` is explicitly set
        // so references to a virtual collection should not be stored in the database
        // But we allow virtual collections to refer to other virtual collections because they are not stored in the database
        if (thisEnd.attr && !thisEnd.model.options.virtualDataSource && otherEnds.some(e => e.model.options.virtualDataSource && !e.model.options.virtualDataSource.hasStableIds)) {
            throw new Error('Non-virtual collection cannot have a dominant relation to a virtual collection without stable IDs');
        }
    }
    /**
     * Gets the alias of this relation, or `undefined`
     * if this end of the relation is not dominant.
     */
    get alias() {
        var _a;
        return (_a = this.thisEnd.attr) === null || _a === void 0 ? void 0 : _a.alias;
    }
    /**
     * Finds references to the related models on the given object.
     * The related models are not necessarily fetched.
     */
    async findRelated(ref, data, transaction) {
        const { attr } = this.thisEnd;
        const related = attr
            ? this._getRefInfo(data, attr)
            : await this._queryRelated(ref, transaction, false);
        return related.map(r => r.ref);
    }
    /**
     * Updates the the related models on the given object.
     */
    async update(ref, prevData, newData, editMode, transaction) {
        const { attr: thisAttr } = this.thisEnd;
        if (thisAttr) {
            // This end is dominant
            // So we know all the other ends directly without querying
            // Update operations will not touch keys that don't exist
            // If the data doesn't have the key, then don't update the relation because we aren't touching it
            // If newData is undefined then we are deleting and we do need to update the relation
            if ((editMode === 'update') && newData && (_.get(newData, thisAttr.alias) === undefined)) {
                return;
            }
            const prevValues = this._getRefInfo(prevData, thisAttr);
            const newValues = this._getRefInfo(newData, thisAttr);
            // Set the value stored in this document appropriately
            this._setThis(newData, thisAttr, newValues.map(v => this._makeRefToOther(v.ref, thisAttr)));
            // Set the value stored in the references documents appropriately
            const removed = _.differenceWith(prevValues, newValues, (a, b) => (0, reference_1.isEqualHandlingRef)(a.ref, b.ref));
            const added = _.differenceWith(newValues, prevValues, (a, b) => (0, reference_1.isEqualHandlingRef)(a.ref, b.ref));
            const related = [
                ...removed.map(info => ({ info, set: false })),
                ...added.map(info => ({ info, set: true })),
            ];
            await this._setAllRelated(related, ref, transaction);
        }
        else {
            // I.e. thisAttr == null (meaning this end isn't dominant)
            if (!newData) {
                // This end is being deleted and it is not dominant
                // so we need to search for the dangling references existing on other models
                const related = await this._queryRelated(ref, transaction, true);
                await this._setAllRelated(related.map(info => ({ info, set: false })), ref, transaction);
            }
            else {
                // This end isn't dominant
                // But it isn't being deleted so there is 
                // no action required on the other side
            }
        }
    }
    /**
     * Populates the related models onto the given object for this relation.
     */
    async populateRelated(ref, data, transaction) {
        const { attr } = this.thisEnd;
        // This could be a partial data update
        // If the attribute value does not exist in the data, then we ignore it
        if (attr && (_.get(data, attr.alias) !== undefined)) {
            const related = await this.findRelated(ref, data, transaction);
            const results = related.length ? await transaction.getNonAtomic(related) : [];
            const values = (0, map_not_null_1.mapNotNull)(results, snap => {
                const data = snap.data();
                if (!data) {
                    // TODO:
                    // Should we throw an error if the reference can't be found or just silently omit it?
                    strapi.log.warn(`Could not populate the reference "${snap.ref.path}" because it no longer exists. ` +
                        'This may be because the database has been modified outside of Strapi, or there is a bug in the Firestore connector.');
                }
                return data;
            });
            // The values will be correctly coerced
            // into and array or single value by the method below
            this._setThis(data, attr, values);
        }
    }
    get _singleOtherEnd() {
        return (this.otherEnds.length === 1) ? this.otherEnds[0] : undefined;
    }
    /**
     * Creates an appropriate `ReferenceShape` to store in the documents
     * at the other end, properly handling polymorphic references.
     *
     * @param ref The reference to this
     * @param otherAttr Attribute info of the other end (which refers to this)
     */
    _makeRefToThis(ref, otherAttr) {
        if (otherAttr.isMorph && !(ref instanceof morph_reference_1.MorphReference)) {
            const { attr } = this.thisEnd;
            if (!attr && otherAttr.filter) {
                throw new Error('Polymorphic reference does not have the required information');
            }
            if ((ref instanceof normal_reference_1.NormalReference) || (ref instanceof deep_reference_1.DeepReference)) {
                ref = new morph_reference_1.MorphReference(ref, attr ? attr.alias : null);
            }
            else {
                throw new Error(`Unknown type of reference: ${ref}`);
            }
        }
        return ref;
    }
    /**
     * Checks the `Reference` to store in this document,
     * properly handling polymorphic references.
     *
     * @param otherRef The reference to the other end
     */
    _makeRefToOther(otherRef, thisAttr) {
        if (otherRef) {
            if (thisAttr.isMorph && !(otherRef instanceof morph_reference_1.MorphReference)) {
                // The reference would have been coerced to an instance of MorphReference
                // only if it was an object with the required info
                throw new Error('Polymorphic reference does not have the required information');
            }
            else {
                return otherRef;
            }
        }
        return null;
    }
    _setThis(data, { alias, isArray }, value) {
        if (data) {
            if (isArray) {
                const val = value ? _.castArray(value) : [];
                _.set(data, alias, val);
            }
            else {
                const val = value ? (Array.isArray(value) ? value[0] || null : value) : null;
                _.set(data, alias, val);
            }
        }
    }
    async _setAllRelated(refs, thisRef, transaction) {
        refs = refs.filter(r => r.info.attr);
        // Batch-get all the references that we need to fetch
        // I.e. the ones inside component arrays that required manual manipulation
        const toGet = [];
        const infos = new Array(refs.length);
        for (let i = 0; i < refs.length; i++) {
            const { info, set } = refs[i];
            // Filter to those that have a dominant other end
            if (info.attr) {
                infos[i] = {
                    attr: info.attr,
                    ref: info.ref,
                    thisRefValue: info.thisRefValue,
                    set,
                };
                // Set aside to fetch this relation
                if (info.attr.isMeta) {
                    infos[i].snapIndex = toGet.length;
                    toGet.push(info.ref);
                }
            }
        }
        const snaps = toGet.length ? await transaction.getAtomic(toGet) : [];
        // Perform all the write operations on the relations
        await Promise.all(infos.map(async (info) => {
            if (info) {
                const data = info.snapIndex !== undefined ? snaps[info.snapIndex].data() : undefined;
                const thisRefValue = info.thisRefValue || this._makeRefToThis(thisRef, info.attr);
                await this._setRelated(info.ref, info.attr, data, thisRefValue, info.set, transaction);
            }
        }));
    }
    async _setRelated(ref, attr, prevData, thisRefValue, set, transaction) {
        const value = set
            ? (attr.isArray ? field_operation_1.FieldOperation.arrayUnion(thisRefValue) : thisRefValue)
            : (attr.isArray ? field_operation_1.FieldOperation.arrayRemove(thisRefValue) : null);
        if (attr.isMeta) {
            if (!prevData) {
                // Relation no longer exists, do not update
                return;
            }
            const { componentAlias, parentAlias } = attr.actualAlias;
            // The attribute is a metadata map for an array of components
            // This requires special handling
            // We need to atomically fetch and process the data then update
            // Extract a new object with only the fields that are being updated
            const newData = {};
            const components = _.get(prevData, parentAlias);
            _.set(newData, parentAlias, components);
            for (const component of _.castArray(components)) {
                if (component) {
                    field_operation_1.FieldOperation.apply(component, componentAlias, value);
                }
            }
            await transaction.update(ref, newData, { updateRelations: false });
        }
        else {
            // TODO: Safely handle relations that no longer exist
            await transaction.update(ref, { [attr.alias]: value }, { updateRelations: false });
        }
    }
    async _queryRelated(ref, transaction, atomic, otherEnds = this.otherEnds) {
        const snaps = otherEnds.map(async (otherEnd) => {
            const { model, attr, parentModels } = otherEnd;
            if (parentModels && parentModels.length) {
                // Find instances of the parent document containing
                // a component instance that references this
                return await this._queryRelated(ref, transaction, atomic, parentModels);
            }
            if (attr) {
                // The refValue will be coerced appropriately
                // by the model that is performing the query
                const refValue = this._makeRefToThis(ref, attr);
                const operator = attr.isArray ? 'array-contains' : '==';
                let q = model.db.where({ field: attr.alias, operator, value: refValue });
                if (model.options.maxQuerySize) {
                    q = q.limit(model.options.maxQuerySize);
                }
                const snap = atomic
                    ? await transaction.getAtomic(q)
                    : await transaction.getNonAtomic(q);
                return snap.docs.map(d => makeRefInfo(otherEnd, d.ref, refValue));
            }
            else {
                return [];
            }
        });
        return (await Promise.all(snaps)).flat();
    }
    _getRefInfo(data, thisAttr) {
        return (0, map_not_null_1.mapNotNull)(_.castArray(_.get(data, thisAttr.alias) || []), v => this._getSingleRefInfo(v));
    }
    _getSingleRefInfo(ref) {
        let other = this._singleOtherEnd;
        if (ref) {
            if (!(ref instanceof reference_1.Reference)) {
                throw new Error('Value is not an instance of Reference. Data must be coerced before updating relations.');
            }
            if (!other) {
                // Find the end which this reference relates to
                other = this.otherEnds.find(({ model }) => model.db.path === ref.parent.path);
                if (!other) {
                    throw new status_error_1.StatusError(`Reference "${ref.path}" does not refer to any of the available models: `
                        + this.otherEnds.map(e => `"${e.model.uid}"`).join(', '), 400);
                }
            }
            return makeRefInfo(other, ref, undefined);
        }
        return null;
    }
}
exports.RelationHandler = RelationHandler;
function makeRefInfo(info, ref, thisRefValue) {
    return {
        ...info,
        ref,
        thisRefValue,
    };
}
