"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.makeTransactionRunner = void 0;
const readwrite_transaction_1 = require("../db/readwrite-transaction");
const p_queue_1 = require("p-queue");
const readonly_transaction_1 = require("../db/readonly-transaction");
/**
 * Makes a function that runs a transaction.
 * If the connector is using an emulator then the return runner queues transactions
 * with a maximum concurrency limit.
 */
function makeTransactionRunner(firestore, options, connectorOptions) {
    const { useEmulator, logTransactionStats } = connectorOptions;
    const isVirtual = options.virtualDataSource != null;
    const normalRunner = async (fn, opts) => {
        const isReadOnly = isVirtual || (opts && opts.readOnly);
        let hooks;
        let result;
        if (isReadOnly) {
            // Always use read-only transactions for virtual collections
            // The only scenario where a virtual collection may want to perform a Firestore write is if it has
            // a dominant relation to a non-virtual collection. However, because of the (potentially) transient nature of
            // references to virtual collections, dominant relations to a virtual collection are not supported.
            // Don't log stats for virtual collections
            const trans = new readonly_transaction_1.ReadOnlyTransaction(firestore, logTransactionStats && !isVirtual);
            result = await fn(trans);
            hooks = trans.successHooks;
            await trans.commit();
        }
        else {
            let attempt = 0;
            const r = await firestore.runTransaction(async (trans) => {
                if ((attempt > 0) && useEmulator) {
                    // Random back-off for contested transactions only when running on the emulator
                    // The production server has deadlock avoidance but the emulator currently doesn't
                    // See https://github.com/firebase/firebase-tools/issues/1629#issuecomment-525464351
                    // See https://github.com/firebase/firebase-tools/issues/2452
                    const ms = Math.random() * 5000;
                    strapi.log.warn(`There is contention on a document and the Firestore emulator is getting deadlocked. Waiting ${ms.toFixed(0)}ms.`);
                    await new Promise(resolve => setTimeout(resolve, ms));
                }
                const wrapper = new readwrite_transaction_1.ReadWriteTransaction(firestore, trans, logTransactionStats, ++attempt);
                const result = await fn(wrapper);
                await wrapper.commit();
                return { result, hooks: wrapper.successHooks };
            }, { maxAttempts: opts === null || opts === void 0 ? void 0 : opts.maxAttempts });
            result = r.result;
            hooks = r.hooks;
        }
        // Run the hooks
        if (hooks.length) {
            await Promise.all(hooks.map(async (hook) => {
                try {
                    await hook();
                }
                catch (err) {
                    strapi.log.error(`Error running transaction success hook: ${err.message}`, { err });
                }
            }));
        }
        return result;
    };
    // Virtual option overrides flatten option
    if (options.flatten && !isVirtual) {
        const queue = new p_queue_1.default({ concurrency: 1 });
        return async (fn, opts) => {
            if (opts && opts.readOnly) {
                // Read-only transactions can succeed concurrently because they don't lock any documents
                // and will not contend with any other transactions
                return normalRunner(fn, opts);
            }
            else {
                // When using flattened collections, only one read-write transaction can be executed at a time
                // So we queue them up rather than allowing them to contend 
                // Contention which would introduce 30-second timeout delays on the emulator, and cause unnecessary
                // read and write operations on the production server
                return await queue.add(() => normalRunner(fn));
            }
        };
    }
    else {
        return normalRunner;
    }
}
exports.makeTransactionRunner = makeTransactionRunner;
